/* Copyright (c) 2009-2010 Linuxense Information Systems. 
 */

#include <ldap.h>
#include <time.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <openssl/sha.h>
#include <regex.h>

#include "lib.h"
#include "array.h"
#include "str-sanitize.h"
#include "mail-storage-private.h"
#include "mailbox-list-private.h"
#include "message-date.h"
#include "deliver.h"
#include "smtp-client.h"
#include "istream.h"
#include "hash.h"
#include "ioloop.h"
#include "hostpid.h"
#include "var-expand.h"
#include "duplicate.h"
#include "istream-header-filter.h"
#include "mkdir-parents.h"

#include "lda-group-plugin.h"

#define LDA_GROUP_CONTEXT(obj)  MODULE_CONTEXT(obj, lda_group_storage_module)
#define LDA_GROUP_MAIL_CONTEXT(obj)  MODULE_CONTEXT(obj, lda_group_mail_module)

static MODULE_CONTEXT_DEFINE_INIT(lda_group_mail_module, &mail_module_register);

static MODULE_CONTEXT_DEFINE_INIT(lda_group_storage_module,
                                  &mail_storage_module_register);

static void (*lda_group_next_hook_mail_storage_created)
        (struct mail_storage *storage);

static void (*lda_group_next_hook_mail_storage_created)
        (struct mail_storage *storage);

/*
 * Global variables 
 */
const unsigned char *data;
char group_hash[50];
char username[256];
char domain[256];
char action[10];
char searchBase[20], ldap_host[50], ldap_dn[50], ldap_dnpass[50];

int isgroup( char *ldapHost, int ldapPort, char *loginDN, char *password, char *searchBase, char *uid );
static bool lda_group_send_mail(const char *recipient, const char *sender, const char *content, char *fname, const char *subject);
static bool lda_group_send_forward(const char *recipient,  char *fname);
LDAP * ldap_connect( char *ldapHost, int ldapPort, char *loginDN, char *password );
void ldap_disconnect( LDAP *ld);
void send_moderation( char ** moderators, char *sender, char *nname ,char *text);
void blast( char **member, char *file );
void chomp(char *string);
void create_hash(char *, char *);

void create_hash(char *data, char *hash)
{
	unsigned char obuf[20];
	SHA1( (unsigned char *) data, strlen(data), obuf);
	strcpy(hash,"");
	sprintf(hash, "%02x%02x%02x%02x\
%02x%02x%02x%02x\
%02x%02x%02x%02x\
%02x%02x%02x%02x\
%02x%02x%02x%02x",
obuf[0],obuf[1],
obuf[2],obuf[3],
obuf[4],obuf[5],
obuf[6],obuf[7],
obuf[8],obuf[9],
obuf[10],obuf[11],
obuf[12],obuf[13],
obuf[14],obuf[15],
obuf[16],obuf[17],
obuf[18],obuf[19]);
}

int isgroup( char *ldapHost, int ldapPort, char *loginDN, char *password, char *searchBase, char *uid )
{
	int version, rc, isGroup=0;
	LDAP        *ld;
	LDAPMessage *searchResult, *entry;
	struct timeval timeOut = {10,0};
	char filter[100];
	strcpy(filter, "(&(objectClass=qmailGroup) (uid=");
	strcat(filter, uid);
	strcat(filter, "))");
	/* Set LDAP version to 3 and set connection timeout. */
	version = LDAP_VERSION3;
	ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION, &version);
	ldap_set_option( NULL, LDAP_OPT_NETWORK_TIMEOUT, &timeOut);
	/* Initialize the LDAP session */
	if (( ld = ldap_init( ldapHost, ldapPort )) == NULL) {
		i_info( "ldap initialization failed.");
		return( 0 );
	}
	/* Bind to the server */
	rc = ldap_simple_bind_s( ld, loginDN, password );
	if (rc != LDAP_SUCCESS ){
		ldap_unbind_s ( ld );
		i_info( "ldap bind failed" );
		return( 0 );
	}
	/* Search the directory */
	rc = ldap_search_ext_s( ld, searchBase, LDAP_SCOPE_SUBTREE,
												filter, NULL, 0, NULL, NULL, &timeOut,
												LDAP_NO_LIMIT, &searchResult );
	if ( rc != LDAP_SUCCESS ) {
		i_info( "ldap_search_ext_s: %s", ldap_err2string( rc ));
		ldap_msgfree( searchResult );
		ldap_unbind_s( ld );
		return ( 0 );
	}
	entry   =   ldap_first_entry( ld, searchResult );
	if(entry !=  NULL)
		isGroup = 1;
	ldap_msgfree( searchResult );
	ldap_unbind_s( ld );
	return( isGroup );
}

LDAP * ldap_connect( char *ldapHost, int ldapPort, char *loginDN, 
										char *password )
{
	int version, rc;
	LDAP        *ld;
	struct timeval timeOut = {10,0};
	/* Set LDAP version to 3 and set connection timeout. */
	version = LDAP_VERSION3;
	ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION, &version);
	ldap_set_option( NULL, LDAP_OPT_NETWORK_TIMEOUT, &timeOut);
	/* Initialize the LDAP session */
	if (( ld = ldap_init( ldapHost, ldapPort )) == NULL) {
		i_info( "Cannot connect to %s:%d", ldapHost, ldapPort);		
		return( NULL );
	}
	/* Bind to the server */
	rc = ldap_simple_bind_s( ld, loginDN, password );
	if (rc != LDAP_SUCCESS ){
		i_info( "Cannot bind with  %s&%s", loginDN, password );		
		ldap_unbind_s ( ld );
		return( NULL );
	}
	return( ld );
}

void ldap_disconnect( LDAP *ld) {

	ldap_unbind_s( ld );
}

void send_moderation( char ** moderators, char *sender, char *nname, char *text )
{
	int i;
	for ( i = 0; moderators[i] != NULL; i++ ) {
		i_info(" Sending Moderation request to %s.", moderators[i]);
		lda_group_send_mail( moderators[i], sender, text, nname, "Moderation Required");
	}
}

void blast( char **member, char *file )
{
	int i;
	for ( i = 0; member[i] != NULL; i++ ) {
		i_info( " Sending Mail to the group member <%s>.", member[i]);
		lda_group_send_forward( member[i], file);
	}
}

void chomp(char *string) {
	
	int i;
	for( i=0; (string[i] != '\0' && string[i] != '\n');i++);
	string[i] = '\0';
}

/*
 * Parse the conf file to get ldap credentials

int get_ldap_login_details( char *ldap_host, char *ldap_dn,
							char *ldap_dnpass, char *searchBase )
{
	FILE *fd;
	char line[200], key[50], value[50];
	regex_t preg;
	regmatch_t pmatch[3];
	size_t rm;
	char pom[1024];
	char conf[] = "/usr/local/etc/dovecot-ldap.conf";
	fd = fopen( conf, "r");
	if (fd) {
		while(!feof(fd)) {
			fgets( line, 200, fd);
			regcomp (&preg, "^(hosts|dn|dnpass|base)[ ]*=[ ]*([^ ]*)", REG_EXTENDED);
			rm = regexec (&preg, line, 3, pmatch, 0);
			if( rm == 0) {
				strncpy (pom, line + pmatch[1].rm_so, pmatch[1].rm_eo - pmatch[1].rm_so);
				pom[pmatch[1].rm_eo - pmatch[1].rm_so] = '\0';
				strncpy( key, pom, strlen(pom)+1 );

				strncpy (pom, line + pmatch[2].rm_so, pmatch[2].rm_eo - pmatch[2].rm_so);
				pom[pmatch[2].rm_eo - pmatch[2].rm_so] = '\0';
				strncpy( value, pom, strlen(pom)+1 );

				i_info( "Conf: %s = %s", key, value);

				if( strcmp(key, "hosts") == 0) {
					strncpy(ldap_host, value, strlen(value)+1 );
					chomp(ldap_host);
				}else if( strcmp(key, "dn") == 0) {
				    strncpy(ldap_dn, value, strlen(value)+1 );
					chomp(ldap_dn);
				}else if( strcmp(key, "dnpass") == 0) {
				    strncpy(ldap_dnpass, value, strlen(value)+1 );
					chomp(ldap_dnpass);
				}else if( strcmp(key, "base") == 0) {
					strncpy(searchBase, value, strlen(value)+1 );
					chomp(searchBase);
				}
			}
		}
		fclose(fd);
	}else {
		i_info( "Conf: Couldn't read conf %s", conf);
	}
	return 0;
}*/

/*
 * Plugin implementation
 */

static struct mail *
lda_group_maildir_mail_alloc(struct mailbox_transaction_context *t,
                        enum mail_fetch_field wanted_fields,
                        struct mailbox_header_lookup_ctx *wanted_headers)
{
 	union mailbox_module_context *zbox = LDA_GROUP_CONTEXT(t->box);
	union mail_module_context *zmail;
	struct mail *_mail;
	struct mail_private *mail;
	char sender[256];
	char group_id[256];
	char recipient[256];
	char fname[200];
	char nname[200];
	FILE *fd;
	
	_mail = zbox->super.mail_alloc(t, wanted_fields, wanted_headers);
	mail = (struct mail_private *)_mail;
	zmail = p_new(mail->pool, union mail_module_context, 1);
	zmail->super = mail->v;
	sprintf( group_id, "%s@%s", username, domain );

	int version, rc=0, i;
	BerElement  *ber;
	LDAP        *ld;
	LDAPMessage *searchResult;
	struct timeval timeOut = {10,0};
	char *attribute , filter[300];
	char *attrs[]= { "objectClass", "senderconfirm", "membersonly", "rfc822member", "rfc822sender", 
									"rfc822moderator", "confirmtext", "moderatortext", NULL };
	char **senderconfirm = NULL , **member=NULL, **moderators=NULL;
	char **confirmtext = NULL, **moderatortext = NULL, **senders = NULL, **membersonly = NULL;
	strcpy(filter, "(&(objectClass=qmailGroup) (uid=");
	strcat(filter, group_id);
	strcat(filter, ")(!(|(accountstatus=deleted)(accountstatus=disabled))))");
	/* Set LDAP version to 3 and set connection timeout. */
	version = LDAP_VERSION3;
	ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION, &version);
	ldap_set_option( NULL, LDAP_OPT_NETWORK_TIMEOUT, &timeOut);
	//get_ldap_login_details( ldap_host, ldap_dn, ldap_dnpass, searchBase );
	ld = ldap_connect( ldap_host, 389 , ldap_dn, ldap_dnpass );
	
	if( ld == NULL ) {
		i_info( "Coldn't connect to ldap server " );
	}else {
		rc = ldap_search_ext_s(
				ld, (char *)searchBase, LDAP_SCOPE_SUBTREE,
				filter, attrs, 0, NULL, NULL, &timeOut,
				LDAP_NO_LIMIT, &searchResult );
	}
	if( rc != LDAP_SUCCESS )
	{   
		i_info( "ldap_search_ext_s: %s", ldap_err2string( rc ) );
		ldap_msgfree( searchResult );
		goto cleanup;
	}

	for( attribute = ldap_first_attribute( ld, searchResult, &ber );
			attribute != NULL;
			attribute = ldap_next_attribute( ld, searchResult, ber ) )
	{
		/* Save the member attribute.  This may also show up as uniqueMember. */
		if (strcmp(attribute, "senderconfirm") == 0)
		{   
			senderconfirm = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "membersonly") == 0 )
		{
			membersonly = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "rfc822member") == 0 )
		{
			member = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "rfc822moderator") == 0)
		{
			moderators = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "rfc822sender") == 0 )
		{
			senders = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "confirmtext") == 0)
		{
			confirmtext = ldap_get_values( ld, searchResult, attribute );
		}
		else if (strcmp(attribute, "moderatortext") == 0)
		{
			moderatortext = ldap_get_values( ld, searchResult, attribute );
		}
		ldap_memfree( attribute );
	}
	ber_free( ber, 0 );
	ldap_msgfree( searchResult );
	mkdir_parents(getenv("HOME"), S_IRWXU);

	if(strcmp(action, "") == 0) {
		if( membersonly != NULL && strcmp(membersonly[0], "TRUE")  == 0) {
			strcpy(sender, deliver_get_return_address(_mail));
			rc = 0;
			if (senders!=NULL) {
				for (i=0; senders[i]; i++){
					if( strcmp(senders[i], sender) == 0) {
						rc = 1;
						break;
					}
				}
			}
			if (moderators!=NULL && rc == 0) {
				for (i=0; moderators[i]; i++){
					if( strcmp(moderators[i], sender) == 0) {
						rc = 1;
						break;
					}
				}
			}
			if(rc == 0) {
				i_info("Sender %s is not allowed to send to %s", sender, group_id);
				sprintf(fname, "%s/.temp%ld", getenv("HOME"), time(NULL));
				sprintf(sender, "%s-noreply@%s", username, domain);
				if (!(fd = fopen(fname, "w"))) {
					i_info("Could not open \"%s\" for output, bailing...", fname);
					exit(EXIT_FAILURE);
				}
				else {
					fprintf(fd, "%s", data);
					fclose(fd);
					strcpy(recipient, deliver_get_return_address(_mail));
					lda_group_send_mail(recipient, sender, "Sorry, you are not allowed to send to this group.", fname, "Mail rejected");
					remove(fname);
				}
				goto cleanup;
			}
		}

		strcpy(sender,"");
		create_hash((char *)data, group_hash);
		i_info("Alloc: [ New Hash: %s ]", group_hash);
		if( senderconfirm != NULL && (strcmp(senderconfirm[0], "TRUE") == 0) ) {
			sprintf(fname, "%s/.%s_c", getenv("HOME"), group_hash);
			sprintf(sender, "%s-confirm-%s@%s", username, group_hash, domain);
			if (!(fd = fopen(fname, "w"))) {
				i_info("Could not open \"%s\" for output, bailing...", fname);
				exit(EXIT_FAILURE);
			}
			else {
				fprintf(fd, "%s", data);
				fclose(fd);
				strcpy(recipient, deliver_get_return_address(_mail));
				lda_group_send_mail(recipient, sender, confirmtext[0], fname, "Confirmation is Required");
				i_info("Confirmation Send to <%s>", recipient);
			}
		}else {
			if( moderators != NULL ) {
				sprintf(fname, "%s/.%s_m", getenv("HOME"), group_hash);
				sprintf(sender, "%s-moderate-%s@%s", username, group_hash, domain);
				if (!(fd = fopen(fname, "w"))) {
					i_info("Could not open \"%s\" for output, bailing...", fname);
					exit(EXIT_FAILURE);
				}
				else {
					fprintf(fd, "%s", data);
					fclose(fd);
					send_moderation( moderators, sender, fname, moderatortext[0]);
				}
			}else {
				sprintf(fname, "%s/.%s", getenv("HOME"), group_hash);
				if (!(fd = fopen(fname, "w"))) {
					i_info("Could not open \"%s\" for output, bailing...", fname);
					exit(EXIT_FAILURE);
				}
				else {
					fprintf(fd, "%s", data);
					fclose(fd);
					if( member != NULL) {
						blast(member, fname);
					}
					remove( fname );
				}
			}
		}
		ldap_value_free( senderconfirm );
		ldap_value_free( moderators );
		ldap_value_free( member );
		ldap_disconnect( ld);
		MODULE_CONTEXT_SET_SELF(mail, lda_group_mail_module, zmail);
		return _mail;
	}
	//Recieved Confirmation from the sender
	else if(strcmp(action, "confirm") == 0) {
		sprintf(fname, "%s/.%s_c", getenv("HOME"), group_hash);
		i_info("Alloc: [ Confirm: Hash: %s ]", group_hash);
		if ((fd = fopen(fname, "r")) == NULL)
		{
			goto cleanup;
		}
		fclose(fd);
		if( moderators != NULL ) {
			sprintf(nname, "%s/.%s_m", getenv("HOME"), group_hash);
			sprintf(sender, "%s-moderate-%s@%s", username, group_hash, domain);
			rename( fname, nname);
			send_moderation( moderators, sender, nname, moderatortext[0]);
		}else {
			if( member != NULL) {
				blast( member, fname);
			}
			remove( fname );
		}
	}
	//Recieved Moderation from the moderator
	else if(strcmp(action, "moderate") == 0) {
		sprintf(fname, "%s/.%s_m", getenv("HOME"), group_hash);
		i_info("Alloc: [ Moderate: Hash: %s ]", group_hash);
		if ((fd = fopen(fname, "r")) == NULL)
        {   
            goto cleanup;
        }
        fclose(fd);
		if( member != NULL) {
			blast(member, fname);
		}
		remove( fname );
	}

cleanup:
	ldap_value_free( senderconfirm );
	ldap_value_free( moderators );
	ldap_value_free( member );
	ldap_disconnect( ld);

	exit(0);
}

static struct mailbox *
lda_group_mailbox_open(struct mail_storage *storage, const char *name,
                      struct istream *input, enum mailbox_open_flags flags)
{
	union mail_storage_module_context *qstorage = LDA_GROUP_CONTEXT(storage);
	struct mailbox *box;
	union mailbox_module_context *lbox;
	struct istream *lda_group_input = NULL;
	size_t size;

	lda_group_input = input;

	box = qstorage->super.mailbox_open(storage, name, input, flags);
	if (box == NULL)
		return NULL;

	lbox = p_new(box->pool, union mailbox_module_context, 1);
	lbox->super = box->v;
	if( strcmp(name, "Dovecot Delivery Mail") == 0) {
		i_stream_read_data(lda_group_input, &data, &size, 1);
		if( size >= 2 ) {
			box->v.mail_alloc = lda_group_maildir_mail_alloc; 
		}
	}

	MODULE_CONTEXT_SET_SELF(box, lda_group_storage_module, lbox);

	return box;
}


static void lda_group_mail_storage_created(struct mail_storage *storage)
{
	union mail_storage_module_context *lstorage;
	lstorage = p_new(storage->pool, union mail_storage_module_context, 1);
	lstorage->super = storage->v;
	char user[256];
	char group_id[100];
	regex_t preg;
	regmatch_t pmatch[6];
	char pom[1024];
	size_t rm;
	if( strcmp(storage->name, "raw") == 0 ) {
		regcomp (&preg, "(^[a-z].*)(@(.*))$", REG_EXTENDED);
		rm = regexec (&preg, storage->ns->user->username, 4, pmatch, 0);

		strncpy( user, storage->ns->user->username, strlen(storage->ns->user->username) );
		strncpy (pom, user + pmatch[1].rm_so, pmatch[1].rm_eo - pmatch[1].rm_so);
		pom[pmatch[1].rm_eo - pmatch[1].rm_so] = '\0';
		strncpy(username, pom, strlen(pom)+1);
		chomp(username);
		strncpy (pom, user + pmatch[3].rm_so, pmatch[3].rm_eo - pmatch[3].rm_so);
		pom[pmatch[3].rm_eo - pmatch[3].rm_so] = '\0';
		strncpy(domain, pom, strlen(pom)+1);
		chomp(username);
		regfree (&preg);
	
		regcomp (&preg, "(^[a-z].*)(-(confirm|moderate|reject)-(.{40}))$", REG_EXTENDED);
		rm = regexec (&preg, username, 5, pmatch, 0);

		if( rm == 0 ) {

			strncpy( user, username, strlen(username) );
			i_info("[Mail storage] Group match found <%s>", username);

			strncpy (pom, user + pmatch[1].rm_so, pmatch[1].rm_eo - pmatch[1].rm_so);
			pom[pmatch[1].rm_eo - pmatch[1].rm_so] = '\0';
			strncpy(username, pom, strlen(pom)+1);
			chomp(username);
			strncpy (pom, user + pmatch[3].rm_so, pmatch[3].rm_eo - pmatch[3].rm_so);
			pom[pmatch[3].rm_eo - pmatch[3].rm_so] = '\0';
			strncpy(action, pom, strlen(pom)+1);
			chomp(action);
			strncpy (pom, user + pmatch[4].rm_so, pmatch[4].rm_eo - pmatch[4].rm_so);
			pom[pmatch[4].rm_eo - pmatch[4].rm_so] = '\0';
			strncpy(group_hash, pom, strlen(pom)+1);
			chomp(group_hash);
		}
		regfree (&preg);
		sprintf(group_id, "%s@%s", username, domain);
		if( isgroup( ldap_host, 389 , ldap_dn, ldap_dnpass , searchBase, group_id) ) {
				storage->v.mailbox_open = lda_group_mailbox_open;
		}
	}

	MODULE_CONTEXT_SET_SELF(storage, lda_group_storage_module, lstorage);

	if (lda_group_next_hook_mail_storage_created != NULL)
		lda_group_next_hook_mail_storage_created(storage);

}


/*
 * Plugin interface
 */

void lda_group_plugin_init(void)
{
	/* Need to find a way to get it from conf */
	strcpy(searchBase, "dc=com");
	strcpy(ldap_host, "localhost");
	strcpy(ldap_dn, "cn=admin,dc=com");
	strcpy(ldap_dnpass,"mystrongpassword");

	/* Hook into the storage created */
	lda_group_next_hook_mail_storage_created = hook_mail_storage_created;
	hook_mail_storage_created =  lda_group_mail_storage_created;
}

void lda_group_plugin_deinit(void)
{
	/* Remove hook */
	hook_mail_storage_created = lda_group_next_hook_mail_storage_created ;
}



/* Send Mail */
static bool lda_group_send_mail(const char *recipient, const char *sender, const char *content, char *fname, const char *subject)
{
	FILE *f, *mail;
	struct smtp_client *smtp_client;
	const char *msgid;
	char ch;
	subject = (char *)str_sanitize(subject, 256);
	mail = fopen(fname, "r");
	/* Open smtp session */
	smtp_client = smtp_client_open(recipient, NULL, &f);
	msgid = deliver_get_new_message_id();
	fprintf(f, "Message-ID: %s\r\n", msgid);
	fprintf(f, "Date: %s\r\n", message_date_create(ioloop_time));
	fprintf(f, "From: Group Mail <%s>\r\n", sender);
	fprintf(f, "To: <%s>\r\n", recipient);
	fprintf(f, "Subject: %s\r\n", subject);
	fprintf(f, "Auto-Submitted: auto-replied (rejected)\r\n");
	fprintf(f, "MIME-Version: 1.0\r\n");
	fprintf(f, "Content-Type: multipart/mixed; boundary=\"%s\"\r\n\r\n", group_hash);
	fprintf(f, "--%s\r\n",group_hash);
	fprintf(f, "Content-Type: text/plain; charset=\"iso-8859-1\"\r\n");
	fprintf(f, "Content-Transfer-Encoding: 8bit\r\n\r\n");
	fprintf(f, "%s\r\n", content);
	fprintf(f, "--%s\r\n", group_hash);
	fprintf(f, "Content-Type: message/rfc822\r\n\r\n");
	while(!feof(mail)) {
		ch = fgetc(mail);
		if(ferror(mail)) {
			i_info("Error reading source file.\n");
			exit(1);
		}
		if(!feof(mail))
			fputc(ch, f);
		if(ferror(f)) {
			i_info("Error writing destination file.\n");
			exit(1);
		}
	}
	fprintf(f, "--%s--\r\n", group_hash);
	return smtp_client_close(smtp_client);
}

/* Forward Mail */
static bool lda_group_send_forward(const char *recipient,  char *fname)
{
	FILE *f, *mail;
	struct smtp_client *smtp_client;
	char ch;
	mail = fopen(fname, "r");
	/* Open smtp session */
	smtp_client = smtp_client_open(recipient, NULL, &f);
	while(!feof(mail)) {
		ch = fgetc(mail);
		if(ferror(mail)) {
			i_info("Error reading source file.\n");
			exit(1);
		}
		if(!feof(mail))
			fputc(ch, f);
		if(ferror(f)) {
			i_info("Error writing destination file.\n");
			exit(1);
		}
	}
	return smtp_client_close(smtp_client);
}
